from __future__ import print_function
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import os
import sys
# add .. to pythonpath
path = os.path.abspath(__file__)
dotdot = os.path.dirname(os.path.dirname(path))
sys.path.append(dotdot)

# add ../util
sys.path.append(os.path.join(dotdot, 'util'))
from an_build_ext import an_build_ext
from setuputils import *

from distutils.core import setup, Extension

import numpy
numpy_inc = numpy.get_include()

def strlist(s, split=' '):
    lst = s.split(split)
    lst = [i.strip() for i in lst]
    lst = [i for i in lst if len(i)]
    return lst

link = ' '.join([os.environ.get('LDFLAGS', ''),
                 os.environ.get('LDLIBS', ''),])
link = strlist(link)
objs = strlist(os.environ.get('SLIB', ''))
inc = strlist(os.environ.get('INC', ''), split='-I')
inc.append('../util') # for util.i
cflags_swig = strlist(os.environ.get('CFLAGS_SWIG', ''))
cflags = strlist(os.environ.get('CFLAGS', ''))

print(('link:', link))
print(('objs:', objs))
print(('inc:', inc))
print(('cflags:', cflags_swig + cflags))

objs = [
    'plotfill.o', 'plotxy.o',
    'plotimage.o', 'plotannotations.o',
    'plotgrid.o', 'plotoutline.o', 'plotindex.o',
    'plotradec.o', 'plothealpix.o', 'plotmatch.o',
    'plotstuff.o', ] + objs

c_module = Extension('_plotstuff_c',
                     sources = [
                         'plotstuff.i',
                         #'plotstuff.c', 'plotfill.c', 'plotxy.c',
                         #'plotimage.c', 'plotannotations.c',
                         #'plotgrid.c', 'plotoutline.c', 'plotindex.c',
                         #'plotradec.c', 'plothealpix.c', 'plotmatch.c',
                         #'matchfile.c', 'matchobj.c',
                         ],
                     include_dirs = [numpy_inc] + inc,
                     depends = objs,
                     extra_objects = objs,
                     extra_compile_args = cflags_swig,
                     extra_link_args=link,
                     swig_opts=['-I'+d for d in inc] + cflags_swig,
    )

setup(cmdclass={'build_ext': an_build_ext},
      name = 'Plotting stuff in python',
      version = '1.0',
      description = 'Just what you need, another plotting package!',
      author = 'Astrometry.net (Dustin Lang)',
      author_email = 'dstn@astro.princeton.edu',
      url = 'http://astrometry.net',
      py_modules = [ 'plotstuff' ],
      ext_modules = [c_module])




from __future__ import print_function
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from run_command import run_command

def get_libs(pkg, required=True):
    (rtn,out,err) = run_command('pkg-config --libs-only-l ' + pkg)
    if rtn:
        if required:
            raise Exception('Failed to find libraries for package ' + pkg)
        else:
            print('Failed to find libraries for (optional) package', pkg)
            return []
    if err and len(err):
        print('pkg-config complained:', err)
    #print 'pkg-config said:', out
    #libs = out.replace('\n', ' ').split(' ')
    libs = out.split()
    libs = [l for l in libs if len(l)]
    # Strip off the leading "-l"
    libs = [l[2:] for l in libs]
    #print 'returning libs:', libs
    return libs

def get_include_dirs(pkg):
    (rtn,out,err) = run_command('pkg-config --cflags-only-I ' + pkg)
    if rtn:
        raise Exception('Failed to find include paths for package ' + pkg)
    if err and len(err):
        print('pkg-config complained:', err)
    dirs = out.split()
    dirs = [l for l in dirs if len(l)]
    # Strip off the leading "-I"
    dirs = [l[2:] for l in dirs]
    #print 'returning include dirs:', dirs
    return dirs

def get_lib_dirs(pkg, required=True):
    (rtn,out,err) = run_command('pkg-config --libs-only-L ' + pkg)
    if rtn:
        if required:
            raise Exception('Failed to find libraries for package ' + pkg)
        else:
            print('Failed to find libraries for (optional) package', pkg)
            return []
    if err and len(err):
        print('pkg-config said:', err)
    libs = out.split()
    libs = [l for l in libs if len(l)]
    # Strip off the leading "-L"
    libs = [l[2:] for l in libs]
    return libs



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
matplotlib.use('Agg')
import pylab as plt
import numpy as np

from astrometry.util.fits import *
from astrometry.util.util import *
from astrometry.util.run_command import *
from astrometry.util.plotutils import *

# Create test file with a grid of RA,Dec points.

r = np.arange(0, 360, 1)
d = np.arange(-90, 90.05, 1)

R,D = np.meshgrid(r, d)

T = tabledata()
T.ra  = R.ravel()
T.dec = D.ravel()
fn = 'test-hpsplit-in1.fits'
T.writeto(fn)

splitpat = 'test-hpsplit-1-%02i.fits'
cmd = 'hpsplit -o %s -n 1 -m 10 %s' % (splitpat, fn)
rtn,out,err = run_command(cmd)
assert(rtn == 0)

ps = PlotSequence('test_hpsplit')
for hp in range(12):
    T = fits_table(splitpat % hp)
    print('Healpix', hp, 'got', len(T))
    plt.clf()
    plothist(T.ra, T.dec, 360, range=((0,360),(-90,90)))
    vv = np.linspace(0, 1, 100)
    rd = []
    for v in vv:
        rd.append(healpix_to_radecdeg(hp, 1, 0., v))
    for v in vv:
        rd.append(healpix_to_radecdeg(hp, 1, v, 1.))
    for v in reversed(vv):
        rd.append(healpix_to_radecdeg(hp, 1, 1., v))
    for v in reversed(vv):
        rd.append(healpix_to_radecdeg(hp, 1, v, 0.))
    rd = np.array(rd)
    plt.plot(rd[:,0], rd[:,1], '-', color=(0,0.5,1.), lw=2)
    plt.axis([0, 360, -90, 90])
    ps.savefig()






# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
import math
from math import exp
from matplotlib.pylab import imread
from numpy.oldnumeric.functions import zeros, ravel

I=imread('3.png')
I=I[:,:,:3]
(h,w,planes) = I.shape
XY = pyfits.open('16b.fits')[1].data
X = XY.field('X')
Y = XY.field('Y')

psfw = 1.0
stars = zeros((h,w)).astype(float)

for (x,y) in zip(X,Y):
    ix = int(round(x))
    iy = int(round(y))
    for dy in range(-5, 6):
        yy = iy + dy
        if yy < 0 or yy >= h:
            continue
        for dx in range(-5, 6):
            xx = ix + dx
            if xx < 0 or xx >= w:
                continue
            dd = (xx - x)**2 + (yy - y)**2
            stars[yy,xx] += exp(-dd / (2 * psfw**2)) #1./(psfw**2 * 2 * math.pi

#origfrac = 0.5
#maxorig = I.max()
#starfrac = (1.0 - origfrac) + (1.0 - maxorig)
#for p in range(planes):
#    I[:,:,p] = I[:,:,p] * origfrac + stars/stars.max() * starfrac

for p in range(planes):
    I[:,:,p] = I[:,:,p] * 0.7 + stars/stars.max() * 0.8

f=open('out.ppm', 'wb')
f.write('P6 %i %i %i\n' % (w, h, 255))
#for j in range(h):
#    for i in range(w):
#        for p in range(planes):
#            f.write(chr(int(round(I[j,i,p] * 255.0))))
flatI = (I.ravel() * 255.0).round().astype(int)
f.write("".join([chr(min(i,255)) for i in flatI]))
f.close()



#/*
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
#*/
from __future__ import print_function
from __future__ import absolute_import

from .fits import *
import numpy as np
import unittest


class TestSlices(unittest.TestCase):
    def setUp(self):
        T = tabledata()
        T.x1 = np.arange(10).astype(int)
        T.x2 = np.arange(10).astype(float)
        T.x3 = [i for i in range(10)]
        T.x4 = tuple([i for i in range(10)])
        self.assertEqual(len(T), 10)
        T.about()
        self.T = T

    def testCut(self):
        T1 = self.T.copy()
        T1.about()

        T1.cut(np.arange(10) < 4)
        T1.about()
        self.assertEqual(len(T1), 4)

    def testIter(self):
        for i in range(len(self.T)):
            self.T[i].about()
        for t in self.T:
            t.about()

    def testNumpySlicing(self):
        A = np.arange(10)
        print(A[:3])
        print(A[3:6])
        print(A[7:])
        print(A[1:7:2])

        A = list(A)
        print(A[:3])
        print(A[3:6])
        print(A[7:])
        print(A[1:7:2])

        A = np.arange(10)
        A = A[:,np.newaxis].repeat(2, axis=1)
        print(A[:3])
        print(A[3:6])
        print(A[7:])
        print(A[1:7:2])

    def testSlice2(self):
        T1 = self.T[self.T.x1 < 4]
        self.assertEqual(len(T1), 4)
        T1.about()
        self.assertTrue(all(T1.x1 == np.array([0,1,2,3])))
        self.assertEqual(T1.x3, [0,1,2,3])

        T1 = self.T[np.flatnonzero(self.T.x1 < 4)]
        self.assertEqual(len(T1), 4)
        T1.about()
        self.assertTrue(all(T1.x1 == np.array([0,1,2,3])))
        self.assertEqual(T1.x3, [0,1,2,3])

    def testSlice1(self):
        T1 = self.T[:3]
        self.assertEqual(len(T1), 3)
        T1.about()
        self.assertTrue(all(T1.x1 == np.arange(3)))
        self.assertTrue(all(T1.x2 == np.arange(3).astype(float)))
        self.assertEqual(T1.x3, [0,1,2])
        self.assertEqual(T1.x4, (0,1,2))

        T2 = self.T[3:6]
        self.assertEqual(len(T2), 3)
        T2.about()
        self.assertEqual(T2.x3, [3,4,5])
        self.assertEqual(T2.x4, (3,4,5))

        T3 = self.T[7:]
        self.assertEqual(len(T3), 3)
        T3.about()
        self.assertEqual(T3.x3, [7,8,9])

        T4 = self.T[1:7:2]
        self.assertEqual(len(T4), 3)
        T4.about()
        self.assertEqual(T4.x3, [1,3,5])


if __name__ == '__main__':
    unittest.main()






# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import numpy

def write_pnm_to(img, f, maxval=255):
    if len(img.shape) == 1:
        raise RuntimeError('write_pnm: img is one-dimensional: must be 2 or 3.')
    elif len(img.shape) == 2:
        #pnmtype = 'G'
        pnmcode = 5
        (h,w) = img.shape
    elif len(img.shape) == 3:
        (h,w,planes) = img.shape
        #pnmtype = 'P'
        pnmcode = 6
        if planes != 3:
            raise RuntimeError('write_pnm: img must have 3 planes, not %i' % planes)
    else:
        raise RuntimeError('write_pnm: img must have <= 3 dimensions.')

    if img.max() > maxval:
        print('write_pnm: Some pixel values are > maxval (%i): clipping them.' % maxval)
    if img.min() < 0:
        print('write_pnm: Some pixel values are < 0: clipping them.')
    clipped = img.clip(0, maxval)

    maxval = int(maxval)
    if maxval > 65535:
        raise RuntimeError('write_pnm: maxval must be <= 65535')
    if maxval < 0:
        raise RuntimeError('write_pnm: maxval must be positive')

    f.write('P%i %i %i %i ' % (pnmcode, w, h, maxval))
    if maxval <= 255:
        f.write(img.astype(numpy.uint8).data)
    else:
        f.write(img.astype(numpy.uint16).data)


def write_pnm(img, filename, maxval=255):
    f = open(filename, 'wb')
    write_pnm_to(img, f, maxval)
    f.close()




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

import os
import sys

from astrometry.util.shell import shell_escape
from astrometry.util.run_command import run_command

# DEBUG
import logging
def logverb(*msg):
    if (sys.version_info > (3,0)):
        logging.debug(' '.join(msg))
    else:
        logging.debug(' '.join([str(m).decode('latin_1', 'backslashreplace') for m in msg]))

# Returns a list (usually with just one element) of 2-tuples:
#  [ (filetype, detail), (filetype, detail) ]
# eg
#  [ ('Minix filesystem', 'version 2'),
#    ('JPEG image data', 'JFIF standard 1.01') ]
def filetype(fn):
    filecmd = 'file -b -N -L -k %s'

    cmd = filecmd % shell_escape(fn)
    (rtn,out,err) = run_command(cmd)
    if rtn:
        logverb('"file" command failed.  Command: "%s"' % cmd)
        logverb('  ', out)
        logverb('  ', err)
        return None

    out = out.strip()
    logverb('File: "%s"' % out)
    lst = []

    # The "file -r" flag, removed in some Ubuntu versions, used to
    # tell it not to convert non-printable characters to octal.  Without -r,
    # some versions print the string r'\012- ' instead of "\n- ".  Do that
    # manually here.
    out = out.replace(r'\012- ', '\n- ')

    for line in out.split('\n- '):
        if line.endswith('\n-'):
            line = line[:-2]
        if len(line) == 0:
            continue
        p = line.split(', ', 1)
        if len(p) == 2:
            lst.append(tuple(p))
        else:
            lst.append((p[0], ''))
    return lst

# Returns a list (usually with just one element) of filetypes, or None if no filetypes are found:
# eg
#  [ 'Minix filesystem', 'JPEG image data' ]
def filetype_short(fn):
    ft = filetype(fn)
    if ft is None:
        return None
    return [t for (t,nil) in ft]



#! /usr/bin/env python
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function

from astrometry.util.index import *
from astrometry.util.sip import *
from astrometry.util.starutil_numpy import *
from astrometry.util.fits import *
from astrometry.libkd.spherematch import match

from numpy import *

from optparse import OptionParser
import sys

def allcombinations(fronts, backs):
    #print 'fronts', fronts, 'backs', backs
    if len(fronts) == 0:
        return fronts
    if len(backs) == 0:
        return fronts
    newf = []
    for f in fronts:
        for b in backs[0]:
            newf.append(f + [b])
    if len(newf) == 0:
        return []
    return allcombinations(newf, backs[1:])
    #return allcombinations([f + [backs[0]] for f in fronts],
    #                       backs[1:])

def whynot(field, index, qidx, wcs):
    # find index stars in the field.
    (ra,dec) = wcs.get_radec_center()
    rad = arcsec2deg(wcs.get_pixel_scale() * hypot(wcs.get_width(), wcs.get_height()))
    
    (xyz, radec, starinds) = index_search_stars(index, ra, dec, rad)
    print('xyz', xyz.shape)
    print('radec', radec.shape)
    print('starinds', starinds.shape)

    istars = tabledata()
    istars.xyz = xyz
    istars.radec = radec
    istars.starind = starinds

    W,H = wcs.get_width(), wcs.get_height()
    pix = array([wcs.radec2pixelxy(r, d) for (r,d) in radec])
    print('pix', pix.shape)
    # within image bounds
    I = (pix[:,0] > 0) * (pix[:,0] < W) * (pix[:,1] > 0) * (pix[:,1] < H)

    # find nearby pairs of stars...
    nsigma = 3
    pixeljitter = 1
    sigma = hypot(index.index_jitter / wcs.get_pixel_scale(), pixeljitter)
    rad = nsigma * sigma

    fieldxy = vstack((field.x, field.y)).T
    print('field', fieldxy.shape)

    (cinds, cdists) = match(pix, fieldxy, rad)
    print('matches:', cinds.shape)
    #print cdists.shape

    corrs = tabledata()
    corrs.dist = cdists[:,0]
    corrs.star = starinds[cinds[:,0]]
    corrs.field = cinds[:,1]

    allquads = []

    # All quads built from stars in the field...
    for starind in starinds[I]:
        #print qidx, starind
        quads = qidxfile_get_quad_list(qidxfile_addr(qidx), starind)
        #print 'quads:', quads.shape
        allquads.append(quads)

    allquads = unique(hstack(allquads))
    print('%i unique quads touch stars in the field' % len(allquads))

    '''
    "quads" object: all quads that touch index stars in this field.

    .quad : (int) quad index
    .stars : (DQ x int) star indices
    .starsinfield: (DQ x bool) stars in field
    
    '''
    quads = tabledata()
    quads.quad = allquads

    qstars = quadfile_get_stars_for_quads(quadfile_addr(index), quads.quad)
    print('stars in quads:', qstars.shape)
    #print qstars

    quads.stars = qstars

    quads.starsinfield = array([[s in starinds[I] for s in q] for q in qstars])
    #print quads.starsinfield

    allin = quads.starsinfield.min(axis=1)
    #print quads.allin

    quads.allin = allin

    quadsin = quads[allin]
    print('%i quads use stars that are in the field' % (len(quadsin)))
    print('stars:', starinds)

    c_s2f = {}
    for c in corrs: #i in range(len(corrs)):
        #print 'corr', c
        if not c.star in c_s2f:
            c_s2f[c.star] = []
        c_s2f[c.star].append(c.field)
        #c_s2f[corrs.star[i]]
    #print c_s2f
    
    # [[[213], [35], [265], [63]], [[186]], [[]], [[11], [19]], ...]
    # For each quad, for each star in the quad, the list of field stars corresponding.
    fq = []
    for q in quadsin.stars:
        fq1 = []
        for s in q:
            if s in c_s2f:
                fq1.append(c_s2f[s])
            else:
                fq1.append([])
        fq.append(fq1)
    #print fq

    # For each quad, the list of sets of field stars corresponding.
    # [[[213, 35, 265, 63]], [], [],
    fq2 = []
    for q in fq:
        #print 'q:', q
        #ac = allcombinations([q[0]], q[1:])
        #ac = allcombinations([[]], q)
        #print '--> ac:', ac
        fq2.append(allcombinations([[]], q))
    #print fq2

    quadsin.fq2 = fq2

    hasf = array([len(x) > 0 for x in quadsin.fq2])
    #print 'hasf:', hasf

    okquads = quadsin[hasf]

    #forder = argsort([max(x) for x in okquads.fq2])
    #okquads = okquads[forder]

    print('ok quads:', len(okquads))
    print('quad nums:', okquads.quad)
    print('stars:', okquads.stars)
    print('field stars:', okquads.fq2)

    #qmatches = table_data()
    #fqs = []
    #qs = []
    #for okq,fq in zip(okquads, okquads.fq2):
    #    fqs = fqs + fq
    #    qs.append(okq)
    #qmatches.fieldquad = fqs

    qmatches = []
    for okq,fqs in zip(okquads, okquads.fq2):
        for fq in fqs:
            #print 'field stars', fq
            #print 'quad:', okq.quad
            qmatches.append((fq, okq))

    forder = argsort([max(fq) for fq,okq in qmatches])
    #print 'forder', forder

    print()
    print('Quads in the order they will be found')
    for i in forder:
        fq,okq = qmatches[i]
        print()
        print('object', max(fq))
        print('field stars', fq)
        print('quad:', okq.quad)




    # Index stars, by sweep (or "r" mag)...
    # --which quads are they part of

    #skdt = starkd_addr(index)
    istars.sweep = array([startree_get_sweep(index.starkd, si) for si in istars.starind])

    mystars = istars[I]
    order = argsort(mystars.sweep)

    print()
    print()
    print('Index stars:', end=' ')
    for ms in mystars[order]:
        print()
        print('Star', ms.starind, 'sweep', ms.sweep, 'radec', ms.radec)
        print('Field star(s) nearby:', len(c_s2f.get(ms.starind, [])))
        for c in corrs[corrs.star == ms.starind]:
            print('  field star', c.field, 'dist', c.dist, 'pixels')
        myquads = quads[array([(ms.starind in q.stars) for q in quads])]
        print('In %i quads' % len(myquads))
        for q in myquads:
            print('  quad %i: %i stars in the field' % (q.quad, sum(q.starsinfield)))
            if q.allin:
                print('    stars:', q.stars)
                print('    correspondences for stars:', [c_s2f.get(s, None) for s in q.stars])


if __name__ == '__main__':
    parser = OptionParser(usage='%prog <field> <index> <qidx> <wcs>')

    opt, args = parser.parse_args()

    if len(args) != 4:
        parser.print_help()
        sys.exit(-1)
        
    fieldfn = args[0]
    indexfn = args[1]
    qidxfn = args[2]
    wcsfn = args[3]

    log_init(3)

    field = fits_table(fieldfn)
    index = index_load(indexfn, 0, None)
    qidx = qidxfile_open(qidxfn)
    wcs = Sip(filename=wcsfn)
    
    whynot(field, index, qidx, wcs)
    



# This file is part of libkd.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from astrometry.libkd import spherematch_c

from astrometry.util.starutil_numpy import radectoxyz, deg2dist, dist2deg, distsq2deg

import numpy as np

def match_xy(x1,y1, x2,y2, R, **kwargs):
    '''
    Like match_radec, except for plain old 2-D points.
    '''
    I,d = match(np.vstack((x1,y1)).T, np.vstack((x2,y2)).T, R, **kwargs)
    return I[:,0],I[:,1],d
    
# Copied from "celestial.py" by Sjoert van Velzen.
def match_radec(ra1, dec1, ra2, dec2, radius_in_deg, notself=False,
                nearest=False, indexlist=False, count=False):
    '''
    Cross-matches numpy arrays of RA,Dec points.

    Behaves like spherematch.pro of IDL.

    Parameters
    ----------
    ra1, dec1, ra2, dec2 : numpy arrays, or scalars.
        RA,Dec in degrees of points to match.

    radius_in_deg : float
        Search radius in degrees.

    notself : boolean
        If True, avoids returning 'identity' matches;
        ASSUMES that ra1,dec1 == ra2,dec2.

    nearest : boolean
        If True, returns only the nearest match in *(ra2,dec2)*
        for each point in *(ra1,dec1)*.

    indexlist : boolean
        If True, returns a list of length *len(ra1)*, containing *None*
        or a list of ints of matched points in *ra2,dec2*.
        

    Returns
    -------
    m1 : numpy array of integers
        Indices into the *ra1,dec1* arrays of matching points.
    m2 : numpy array of integers
        Same, but for *ra2,dec2*.
    d12 : numpy array, float
        Distance, in degrees, between the matching points.
    '''
    # Convert to coordinates on the unit sphere
    xyz1 = radectoxyz(ra1, dec1)
    #if all(ra1 == ra2) and all(dec1 == dec2):
    if ra1 is ra2 and dec1 is dec2:
        xyz2 = xyz1
    else:
        xyz2 = radectoxyz(ra2, dec2)
    r = deg2dist(radius_in_deg)

    extra = ()
    if nearest:
        X = _nearest_func(xyz2, xyz1, r, notself=notself, count=count)
        if not count:
            (inds,dists2) = X
            I = np.flatnonzero(inds >= 0)
            J = inds[I]
            d = distsq2deg(dists2[I])
        else:
            #print 'X', X
            #(inds,dists2,counts) = X
            J,I,d,counts = X
            extra = (counts,)
            print('I', I.shape, I.dtype)
            print('J', J.shape, J.dtype)
            print('counts', counts.shape, counts.dtype)
    else:
        X = match(xyz1, xyz2, r, notself=notself, indexlist=indexlist)
        if indexlist:
            return X
        (inds,dists) = X
        dist_in_deg = dist2deg(dists)
        I,J = inds[:,0], inds[:,1]
        d = dist_in_deg[:,0]
        
    return (I, J, d) + extra


def cluster_radec(ra, dec, R, singles=False):
    '''
    Finds connected groups of objects in RA,Dec space.

    Returns a list of lists of indices that are connected,
    EXCLUDING singletons.

    If *singles* is *True*, also returns the indices of singletons.
    '''
    I,J,d = match_radec(ra, dec, ra, dec, R, notself=True)

    # 'mgroups' maps each index in a group to a list of the group members
    mgroups = {}
    # 'ugroups' is a list of the unique groups
    ugroups = []
    
    for i,j in zip(I,J):
        # Are both sources already in groups?
        if i in mgroups and j in mgroups:
            # Are they already in the same group?
            if mgroups[i] == mgroups[j]:
                continue
            # merge if they are different;
            # assert(they are disjoint)
            lsti = mgroups[i]
            lstj = mgroups[j]
            merge = lsti + lstj
            for k in merge:
                mgroups[k] = merge

            ugroups.remove(lsti)
            ugroups.remove(lstj)
            ugroups.append(merge)

        elif i in mgroups:
            # Add j to i's group
            lst = mgroups[i]
            lst.append(j)
            mgroups[j] = lst
        elif j in mgroups:
            # Add i to j's group
            lst = mgroups[j]
            lst.append(i)
            mgroups[i] = lst
        else:
            # Create a new group
            lst = [i,j]
            mgroups[i] = lst
            mgroups[j] = lst

            ugroups.append(lst)


    if singles:
        S = np.ones(len(ra), bool)
        for g in ugroups:
            S[np.array(g)] = False
        S = np.flatnonzero(S)
        return ugroups,S

    return ugroups





def _cleaninputs(x1, x2):
    fx1 = x1.astype(np.float64)
    if x2 is x1:
        fx2 = fx1
    else:
        fx2 = x2.astype(np.float64)
    (N1,D1) = fx1.shape
    (N2,D2) = fx2.shape
    if D1 != D2:
        raise ValueError('Arrays must have the same dimensionality')
    return (fx1,fx2)

def _buildtrees(x1, x2):
    (fx1, fx2) = _cleaninputs(x1, x2)
    kd1 = spherematch_c.KdTree(fx1)
    if fx2 is fx1:
        kd2 = kd1
    else:
        kd2 = spherematch_c.KdTree(fx2)
    return (kd1, kd2)

def match(x1, x2, radius, notself=False, permuted=True, indexlist=False):
    '''
    ::

        (indices,dists) = match(x1, x2, radius):

    Or::

        inds = match(x1, x2, radius, indexlist=True):

    Returns the indices (Nx2 int array) and distances (Nx1 float
    array) between points in *x1* and *x2* that are within *radius*
    Euclidean distance of each other.

    *x1* is N1xD and *x2* is N2xD.  *x1* and *x2* can be the same
    array.  Dimensions D above 5-10 will probably not run faster than
    naive.

    Despite the name of this package, the arrays x1 and x2 need not be
    celestial positions; in particular, there is no RA wrapping at 0,
    and no special handling at the poles.  If you want to match
    celestial coordinates like RA,Dec, see the match_radec function.

    If *indexlist* is True, the return value is a python list with one
    element per data point in the first tree; that element is a python
    list containing the indices of points matched in the second tree.

    The *indices* return value has a row for each match; the matched
    points are:
    x1[indices[:,0],:]
    and
    x2[indices[:,1],:]

    This function doesn\'t know about spherical coordinates -- it just
    searches for matches in n-dimensional space.

    >>> from astrometry.util.starutil_numpy import *   
    >>> from astrometry.libkd import spherematch
    >>> # RA,Dec in degrees
    >>> ra1  = array([  0,  1, 2, 3, 4, 359,360])
    >>> dec1 = array([-90,-89,-1, 0, 1,  89, 90])
    >>> # xyz: N x 3 array: unit vectors
    >>> xyz1 = radectoxyz(ra1, dec1)
    >>> ra2  = array([ 45,   1,  4, 4, 4,  0,  1])
    >>> dec2 = array([-89, -88, -1, 0, 2, 89, 89])
    >>> xyz2 = radectoxyz(ra2, dec2)
    >>> # The \'radius\' is now distance between points on the unit sphere --
    >>> # for small angles, this is ~ angular distance in radians.  You can use
    >>> # the function:
    >>> radius_in_deg = 2.
    >>> r = sqrt(deg2distsq(radius_in_deg))
    >>> (inds,dists) = spherematch.match(xyz1, xyz2, r)
    >>> # Now *inds* is an Mx2 array of the matching indices,
    >>> # and *dists* the distances between them:
    >>> #  eg,  sqrt(sum((xyz1[inds[:,0],:] - xyz2[inds[:,1],:])**2, axis=1)) = dists
    >>> print inds
    [[0 0]
     [1 0]
     [1 1]
     [2 2]
     [3 2]
     [3 3]
     [4 3]
     [4 4]
     [5 5]
     [6 5]
     [5 6]
     [6 6]]
    >>> print sqrt(sum((xyz1[inds[:,0],:] - xyz2[inds[:,1],:])**2, axis=1))
    [ 0.01745307  0.01307557  0.01745307  0.0348995   0.02468143  0.01745307
      0.01745307  0.01745307  0.0003046   0.01745307  0.00060917  0.01745307]
    >>> print dists[:,0]
    [ 0.01745307  0.01307557  0.01745307  0.0348995   0.02468143  0.01745307
      0.01745307  0.01745307  0.0003046   0.01745307  0.00060917  0.01745307]
    >>> print vstack((ra1[inds[:,0]], dec1[inds[:,0]], ra2[inds[:,1]], dec2[inds[:,1]])).T
    [[  0 -90  45 -89]
     [  1 -89  45 -89]
     [  1 -89   1 -88]
     [  2  -1   4  -1]
     [  3   0   4  -1]
     [  3   0   4   0]
     [  4   1   4   0]
     [  4   1   4   2]
     [359  89   0  89]
     [360  90   0  89]
     [359  89   1  89]
     [360  90   1  89]]

    Parameters
    ----------
    x1 : numpy array, float, shape N1 x D
        First array of points to match

    x2 : numpy array, float, shape N2 x D
        Second array of points to match

    radius : float
        Scalar Euclidean distance to match
        
    Returns
    -------
    indices : numpy array, integers, shape M x 2, for M matches
        The array of matching indices; *indices[:,0]* are indices in *x1*,
        *indices[:,1]* are indices in *x2*.

    dists : numpy array, floats, length M, for M matches
        The distances between matched points.
        
    If *indexlist* is *True*:

    indices : list of ints of integers
        The list of matching indices.  One list element per *x1* element,
        containing a list of matching indices in *x2*.
        
    '''
    (kd1,kd2) = _buildtrees(x1, x2)
    if indexlist:
        inds = spherematch_c.match2(kd1, kd2, radius, notself, permuted)
    else:
        (inds,dists) = spherematch_c.match(kd1, kd2, radius, notself, permuted)
    if indexlist:
        return inds
    return (inds,dists)

def match_naive(x1, x2, radius, notself=False):
    ''' Does the same thing as match(), but the straight-forward slow
    way.  (Not necessarily the way you\'d do it in python either).
    Not very fair as a speed comparison, but useful to convince
    yourself that match() does the right thing.
    '''
    (fx1, fx2) = _cleaninputs(x1, x2)
    (N1,D1) = x1.shape
    (N2,D2) = x2.shape
    inds = []
    dists = []
    for i1 in range(N1):
        for i2 in range(N2):
            if notself and i1 == i2:
                continue
            d2 = sum((x1[i1,:] - x2[i2,:])**2)
            if d2 < radius**2:
                inds.append((i1,i2))
                dists.append(sqrt(d2))
    inds = array(inds)
    dists = array(dists)
    return (inds,dists)

def nearest(x1, x2, maxradius, notself=False, count=False):
    '''
    For each point in x2, returns the index of the nearest point in x1,
    if there is a point within 'maxradius'.

    (Note, this may be backward from what you want/expect!)
    '''
    (kd1,kd2) = _buildtrees(x1, x2)
    if count:
        X = spherematch_c.nearest2(kd1, kd2, maxradius, notself, count)
    else:
        X = spherematch_c.nearest(kd1, kd2, maxradius, notself)
    return X
_nearest_func = nearest

def tree_build_radec(ra=None, dec=None, xyz=None):
    '''
    Builds a kd-tree given *RA,Dec* or unit-sphere *xyz* coordinates.
    '''
    if ra is not None:
        (N,) = ra.shape
        xyz = np.zeros((N,3)).astype(float)
        xyz[:,2] = np.sin(np.deg2rad(dec))
        cosd = np.cos(np.deg2rad(dec))
        xyz[:,0] = cosd * np.cos(np.deg2rad(ra))
        xyz[:,1] = cosd * np.sin(np.deg2rad(ra))
    kd = spherematch_c.KdTree(xyz)
    return kd

def tree_build(X, nleaf=16, bbox=True, split=False):
    '''
    Builds a kd-tree given a numpy array of Euclidean points.
    
    Parameters
    ----------
    X: numpy array of shape (N,D)
        The points to index.
        
    Returns
    -------
    kd: integer
        kd-tree identifier (address).
    '''
    return spherematch_c.KdTree(X, nleaf=nleaf, bbox=bbox, split=split)

def tree_free(kd):
    '''
    Frees a kd-tree previously created with *tree_build*.
    '''
    print('No need for tree_free')
    pass

def tree_save(kd, fn):
    '''
    Writes a kd-tree to the given filename.
    '''
    print('Deprecated tree_save()')
    return kd.write(fn)
#rtn = spherematch_c.kdtree_write(kd, fn)
#return rtn

def tree_open(fn, treename=None):
    '''
    Reads a kd-tree from the given filename.
    '''
    if treename is None:
        return spherematch_c.KdTree(fn)
    else:
        return spherematch_c.KdTree(fn, treename)

def tree_close(kd):
    '''
    Closes a kd-tree previously opened with *tree_open*.
    '''
    print('No need for tree_close')
    pass

def tree_search(kd, pos, radius, getdists=False, sortdists=False):
    '''
    Searches the given kd-tree for points within *radius* of the given
    position *pos*.
    '''
    #print('Unnecessary call to tree_search(kd, ...); use kd.search(...)')
    return kd.search(pos, radius, int(getdists), int(sortdists))

def tree_search_radec(kd, ra, dec, radius, getdists=False, sortdists=False):
    '''
    ra,dec in degrees
    radius in degrees
    '''
    dec = np.deg2rad(dec)
    cosd = np.cos(dec)
    ra = np.deg2rad(ra)
    pos = np.array([cosd * np.cos(ra), cosd * np.sin(ra), np.sin(dec)])
    rad = deg2dist(radius)
    return tree_search(kd, pos, rad, getdists=getdists, sortdists=sortdists)

def trees_match(kd1, kd2, radius, nearest=False, notself=False,
                permuted=True, count=False):
    '''
    Runs rangesearch or nearest-neighbour matching on given kdtrees.

    'radius' is Euclidean distance.

    If 'nearest'=True, returns the nearest neighbour of each point in "kd1";
    ie, "I" will NOT contain duplicates, but "J" may.

    If 'count'=True, also counts the number of objects within range
    as well as returning the nearest neighbor of each point in "kd1";
    the return value becomes I,J,d,counts , counts a numpy array of ints.

    Returns (I, J, d), where
      I are indices into kd1
      J are indices into kd2
      d are distances-squared
      [counts is number of sources in range]

    >>> import numpy as np
    >>> X = np.array([[1, 2, 3, 6]]).T.astype(float)
    >>> Y = np.array([[1, 4, 4]]).T.astype(float)
    >>> kd1 = tree_build(X)
    >>> kd2 = tree_build(Y)
    >>> I,J,d = trees_match(kd1, kd2, 1.1, nearest=True)
    >>> print I
    [0 1 2]
    >>> print J
    [0 0 2]
    >>> print d
    [  0.  60.  60.]
    >>> I,J,d,count = trees_match(kd1, kd2, 1.1, nearest=True, count=True)
    >>> print I
    [0 1 2]
    >>> print J
    [0 0 2]
    >>> print d
    [  0.  60.  60.]
    >>> print count
    [1 1 2]
    '''
    rtn = None
    if nearest:
        rtn = spherematch_c.nearest2(kd2, kd1, radius, notself, count)
        # J,I,d,[count]
        rtn = (rtn[1], rtn[0], np.sqrt(rtn[2])) + rtn[3:]
        #distsq2deg(rtn[2]),
    else:
        (inds,dists) = spherematch_c.match(kd1, kd2, radius, notself, permuted)
        #d = dist2deg(dists[:,0])
        d = dists[:,0]
        I,J = inds[:,0], inds[:,1]
        rtn = (I,J,d)
    return rtn

def tree_permute(kd, I):
    print('Unnecessary call to tree_permute(kd, I): use kd.permute(I)')
    return kd.permute(I)

def tree_bbox(kd):
    print('Unnecessary call to tree_bbox(kd): use kd.bbox')
    return kd.bbox

def tree_n(kd):
    print('Unnecessary call to tree_n(kd): use kd.n')
    return kd.n

def tree_print(kd):
    print('Unnecessary call to tree_print(kd): use kd.print()')
    kd.print()
    
def tree_data(kd, I):
    print('Unnecessary call to tree_data(kd, I): use kd.get_data(I)')
    return kd.get_data(I)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import numpy as np

def usnob_apply_cuts(X):
    # USNO-B sources (not Tycho-2)
    I = (X.num_detections >= 2)
    # no diffraction spikes
    I = np.logical_and(I, np.logical_not(X.flags[:,0]))
    #X = X[I]
    #print '%i pass USNO-B diffraction spike cut' % len(X)

    if hasattr(X, 'an_diffraction_spike'):
        f = X.an_diffraction_spike
        print('f shape', f.shape)
        #u = unique(f)
        #print 'unique flags:', u
        for j in range(8):
            print('flag', j, 'vals', np.unique(f[:,j]))
        I *= (np.logical_not(f[:,7]) * np.logical_not(f[:,6]))
        #X = X[good]
        #print '%i pass AN diffraction spike cut' % len(X)
    return I

def usnob_compute_average_mags(X):
    # Compute average R and B mags.

    #epoch1 = (X.field_1 > 0)
    #epoch2 = (X.field_3 > 0)
    #nmag = np.where(epoch1, 1, 0) + np.where(epoch2, 1, 0)
    #summag = np.where(epoch1, X.magnitude_1, 0) + np.where(epoch2, X.magnitude_3, 0)
    #X.r_mag = np.where(nmag == 0, 0, summag / nmag)

    # rewritten to (hopefully) use less memory
    X.r_mag = np.zeros(len(X), np.float32)
    # sources that only have a first-epoch measurement
    I = np.flatnonzero((X.field_1 > 0) * (X.field_3 == 0))
    X.r_mag[I] = X.magnitude_1[I]
    # sources that only have a second-epoch measurement
    I = np.flatnonzero((X.field_1 == 0) * (X.field_3 > 0))
    X.r_mag[I] = X.magnitude_3[I]
    # sources that have both
    I = np.flatnonzero((X.field_1 > 0) * (X.field_3 > 0))
    X.r_mag[I] = (X.magnitude_1[I] + X.magnitude_3[I]) / 2.

    # B
    #epoch1 = (X.field_0 > 0)
    #epoch2 = (X.field_2 > 0)
    #nmag = np.where(epoch1, 1, 0) + np.where(epoch2, 1, 0)
    #summag = np.where(epoch1, X.magnitude_0, 0) + np.where(epoch2, X.magnitude_2, 0)
    #X.b_mag = np.where(nmag == 0, 0, summag / nmag)

    X.b_mag = np.zeros(len(X), np.float32)
    #I = np.logical_or(X.field_0 > 0, X.field_2 > 0)
    #X.b_mag[I] = (
    #    ((X.field_0[I] > 0)*X.magnitude_0[I] + (X.field_2[I] > 0)*X.magnitude_2[I]) /
    #    ((X.field_0[I] > 0)*1. + (X.field_2[I] > 0)*1.))

    # sources that only have a first-epoch measurement
    I = np.flatnonzero((X.field_0 > 0) * (X.field_2 == 0))
    X.b_mag[I] = X.magnitude_0[I]
    # sources that only have a second-epoch measurement
    I = np.flatnonzero((X.field_0 == 0) * (X.field_2 > 0))
    X.b_mag[I] = X.magnitude_2[I]
    # sources that have both
    I = np.flatnonzero((X.field_0 > 0) * (X.field_2 > 0))
    X.b_mag[I] = (X.magnitude_0[I] + X.magnitude_2[I]) / 2.



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE


import re
import sys
import socket
import time
from urllib.request import urlopen
from urllib.error import URLError
from urllib.parse import urlencode
from urllib.parse import urlparse, urljoin
from os.path import basename
from optparse import OptionParser

from astrometry.util.starutil_numpy import *
from astrometry.util.file import *

# raw, decw: rectangle size, in arcmin.
# returns list of local filenames to which images were written
def get_usnob_images(ra, dec, raw=15., decw=15., fieldname='',
                     fits=False, wait=True, sources=None, basefn=None,
                     survey='All Surveys', justurls=False):
    if survey is None:
        survey = 'All Surveys'
    if survey == 'poss-i':
        survey = 'POSS-I (103aO, 103aE)'
    if survey == 'poss-ii':
        survey = 'POSS-II (IIIaJ, IIIaF, IV-N)'
    # other survey options include 'SOUTH' and 'AAO-R'

    formvals = {
        'fld': fieldname,
        'ra': ra2hmsstring(ra),
        'dec': dec2dmsstring(dec),
        'rawid': raw,
        'decwid': decw,
        'wunits': 'Minutes',
        'surims': survey,
        'pixflg': 'yes', # images, please
        'whorbl': 'Light Stars/Dark Sky',
        'pixgraph': 'JPEG',
        'cextract': 'rect',
        'ori': 'NW - North Up, East Left',
        'cat': 'USNO B1.0',
        }

    if fits:
        formvals['pixfits'] = 'Yes'
    else:
        formvals['pixfits'] = 'No'
    if sources is None:
        formvals.update({
            'opstars': 'No', # overplot stars?
            'colbits': 'cb_ra',
            })

    else:
        knownsources = ['R1','R2','B1','B2']
        if not sources in knownsources:
            print('Unknown source color', sources)
            print('options:', knownsources)
            return -1
        formvals.update({'opstars': 'Yes',
                         'cat': 'USNO B1.0',
                         'getcat': 'yes',
                         'colbits': 'cb_ra',
                         'clr': 'R1', # overplot color?
                         })

    if basefn is None:
        basefn = 'usnob-%g-%g-' % (ra,dec)

    #queryurl = 'http://www.nofs.navy.mil/cgi-bin/tfch3tI.cgi'
    queryurl = 'http://www.nofs.navy.mil/cgi-bin/tfch4.test.cgi'
    print('submitting form values:')
    for k,v in list(formvals.items()):
        print('  ',k,'=',v)
    print('encoded as:')
    print('  ' + urlencode(formvals))
    print()
    print('waiting for results...')
    socket.setdefaulttimeout(300)
    f = urlopen(queryurl, urlencode(formvals))
    doc = f.read()
    fn = 'res.html'
    write_file(doc, fn)
    m = re.search(r'<A HREF="(.*)">', doc)
    if not m:
        raise RuntimeError('Failed to parse results page: server output written to file ' + fn)
    resurl = m.group(1)
    print('result url', resurl)

    if not wait:
        return resurl
    fns = []
    # keep hitting the url... it may be 404 for a while.
    while True:
        print('requesting result url', resurl)
        res = urlopen(resurl)
        #print 'result info:', res.info()
        print()
        doc = res.read()
        write_file(doc, 'res2.html')

        mjpeg = re.findall(r'<A HREF="([a-zA-z0-9/_.]*?)  ">jpg  </A>', doc)
        print('got jpeg matches:', mjpeg)
        print()
        if not len(mjpeg):
            raise Exception('no jpeg results found')
        if fits:
            mfits = re.findall(r'<A HREF="(.*?)">FITS</A>', doc)
            print('got fits matches:', mfits)
            print()
            if not len(mfits):
                raise Exception('no fits results found')
        else:
            mfits = []

        if justurls:
            if fits:
                return ([urljoin(resurl, url) for url in mjpeg],
                        [urljoin(resurl, url) for url in mfits])
            else:
                return [urljoin(resurl, url) for url in mjpeg]

        for (urls,suffix) in [(mjpeg,''), (mfits,'.fits')]:
            for url in urls:
                print('retrieving image', url)
                fullurl = urljoin(resurl, url)
                print('retrieving url', fullurl)
                fn = basefn + basename(url) + suffix
                for trynum in range(100):
                    try:
                        res = urlopen(fullurl)
                        print()
                        print('writing to file', fn)
                        write_file(res.read(), fn)
                        break
                    except Exception as e:
                        print('Error:', e)
                        print('Retrying...')
                        time.sleep(1)
                fns.append(fn)
        break
    return fns


if __name__ == '__main__':
    parser = OptionParser(usage='%prog [options] <ra> <dec>')

    parser.add_option('--r1', '--R1', dest='red1', help='overplot first-epoch red sources', action='store_true')
    parser.add_option('--r2', '--R2', dest='red2', help='overplot second-epoch red sources', action='store_true')
    parser.add_option('-b', dest='big', help='30x30 arcmin (default: 15x15)', action='store_true')
    parser.add_option('-f', '--fits', dest='fits', help='Get FITS image too', action='store_true')
    parser.add_option('-n', dest='fieldname', help='fieldname (default: NoName)')

    parser.set_defaults(red1=False, red2=False, big=None, fieldname=None, fits=False)

    (opt, args) = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        print()
        print('Got extra arguments:', args)
        sys.exit(-1)

    sources = None
    if opt.red1:
        sources = 'R1'
    elif opt.red2:
        sources = 'R2'

    sz = 15.
    if opt.big:
        sz = 30.

    # parse RA,Dec.
    ra = float(args[0])
    dec = float(args[1])

    args = {}
    if opt.fieldname:
        args['fieldname'] = opt.fieldname
    if opt.fits:
        args['fits'] = True
    get_usnob_images(ra, dec, sources=sources, raw=sz, decw=sz, **args)



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import subprocess
import os
import select
import sys
from subprocess import PIPE
import sys

py3 = (sys.version_info[0] >= 3)

# Returns (rtn, out, err)
def run_command(cmd, timeout=None, callback=None, stdindata=None,
                tostring=True):
    """
    Run a command and return the text written to stdout and stderr, plus
    the return value.

    In python3, if *tostring* is True, the output and error streams
    will be converted to unicode, otherwise will be returned as bytes.
    
    Returns: (int return value, string out, string err)
    """
    child = subprocess.Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)
    (fin, fout, ferr) = (child.stdin, child.stdout, child.stderr)

    stdin = fin.fileno()
    stdout = fout.fileno()
    stderr = ferr.fileno()
    outdata = []
    errdata = []
    ineof = outeof = erreof = False
    block = 1024
    while True:
        readers = []
        writers = []
        if not ineof: writers.append(stdin)
        if not outeof: readers.append(stdout)
        if not erreof: readers.append(stderr)
        if not len(readers):
            break
        (ready_readers, ready_writers, _) = select.select(readers, writers, [], timeout)
        if stdin in ready_writers and stdindata:
            bytes_written = os.write(stdin, stdindata[:block])
            stdindata = stdindata[bytes_written:]
            if not stdindata:
                fin.close()
                ineof = True
        if stdout in ready_readers:
            outchunk = os.read(stdout, block)
            if len(outchunk) == 0:
                outeof = True
            outdata.append(outchunk)
        if stderr in ready_readers:
            errchunk = os.read(stderr, block)
            if len(errchunk) == 0:
                erreof = True
            errdata.append(errchunk)
        if callback:
            callback()
    fout.close()
    ferr.close()
    w = child.wait()
    if py3:
        out = b''.join(outdata)
        err = b''.join(errdata)
        if tostring:
            out = out.decode()
            err = err.decode()
    else:
        out = ''.join(outdata)
        err = ''.join(errdata)

    if not os.WIFEXITED(w):
        return (-100, out, err)
    rtn = os.WEXITSTATUS(w)
    return (rtn, out, err)



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import

from astrometry.util.fits import *
from astrometry.util.starutil_numpy import *
from astrometry.util.find_data_file import *
from os.path import basename,dirname
import numpy as np

def get_photoobj_filename(photoobjdir, rr, run, camcol, field):
    fn = os.path.join(photoobjdir, rr, '%i'%run, '%i'%camcol,
                      'photoObj-%06i-%i-%04i.fits' % (run, camcol, field))
    return fn


def read_photoobjs_in_wcs(wcs, margin, cols=None,
                          cutToPrimary=True,
                          wfn=None,
                          sdss=None):
    '''
    Read photoObjs that are inside the given 'wcs', plus 'margin' in degrees.
    '''
    import logging
    log = logging.getLogger('read_photoobjs_in_wcs')

    ra,dec = wcs.radec_center()
    rad = wcs.radius()
    rad += np.hypot(14., 10.) / 2 / 60.
    # a little extra margin
    rad += margin

    if sdss is None:
        from astrometry.sdss import DR9
        sdss = DR9()
    if wfn is None:
        wfn = sdss.filenames.get('window_flist', None)
        if wfn is None:
            wfn = 'window_flist.fits'
    if not os.path.exists(wfn):
        print('File does not exist:', wfn, '; downloading...')
        wfn = sdss.retrieve('window_flist', None, None, None, rerun='xxx')
        print('Retrieved', wfn)
    else:
        print('Using', wfn)

    print('Searching for run,camcol,fields with radius', rad, 'deg')
    RCF = radec_to_sdss_rcf(ra, dec, radius=rad*60., tablefn=wfn)
    log.debug('Found %i fields possibly in range' % len(RCF))

    pixmargin = margin * 3600. / wcs.pixel_scale()
    W,H = wcs.get_width(), wcs.get_height()
    
    TT = []
    for run,camcol,field,r,d in RCF:
        log.debug('RCF %i/%i/%i' % (run, camcol, field))
        rr = sdss.get_rerun(run, field=field)
        if rr in [None, '157']:
            log.debug('Rerun 157')
            continue

        fn = sdss.retrieve('photoObj', run, camcol, field, rerun=rr)
        #fn = get_photoobj_filename(rr, run, camcol, field)

        T = fits_table(fn, columns=cols)
        if T is None:
            log.debug('read 0 from %s' % fn)
            continue
        log.debug('read %i from %s' % (len(T), fn))

        # while we're reading it, record its length for later...
        #get_photoobj_length(rr, run, camcol, field)

        ok,x,y = wcs.radec2pixelxy(T.ra, T.dec)
        x -= 1
        y -= 1
        T.cut((x > -pixmargin) * (x < (W + pixmargin)) *
              (y > -pixmargin) * (y < (H + pixmargin)))
        if cutToPrimary:
            T.cut((T.resolve_status & 256) > 0)
            log.debug('cut to %i within target area and PRIMARY.' % len(T))
        else:
            log.debug('cut to %i within target area.' % len(T))
        if len(T) == 0:
            continue
        TT.append(T)
    if not len(TT):
        return None
    T = merge_tables(TT)
    return T


class RaDecToRcf(object):
    def __init__(self, tablefn=None):
        self.kd = None
        self.sdssxyz = None
        if tablefn is None:
            tablefn = find_data_file('dr7fields.fits')
        self.tab = fits_table(tablefn)
        if self.tab is None:
            raise Exception('Failed to read table of SDSS fields from file: "' + str(tablefn) + '"')

    def __call__(self, ra, dec, spherematch=True, radius=0, contains=False):
        T = self.tab
        # HACK - magic 13x9 +1 arcmin.
        if radius == 0:
            radius = sqrt(14.**2 + 10.**2)/2.
        d2 = arcmin2distsq(radius)
        if self.sdssxyz is None:
            self.sdssxyz = radectoxyz(T.ra, T.dec)
        if not spherematch:
            rcfs = []
            for r,d in broadcast(ra,dec):
                xyz = radectoxyz(r,d)
                dist2s = sum((xyz - self.sdssxyz)**2, axis=1)
                I = flatnonzero(dist2s < d2)
                rcfs.append(zip(T[I].run, T[I].camcol,
                                T[I].field, T[I].ra, T[I].dec))
        else:
            from astrometry.libkd import spherematch
            if self.kd is None:
                self.kd = spherematch.tree_build(self.sdssxyz)
            rds = array([x for x in broadcast(ra,dec)])
            xyz = radectoxyz(rds[:,0], rds[:,1]).astype(double)
            kd2 = spherematch.tree_build(xyz)
            notself = False
            inds,D = spherematch.trees_match(self.kd, kd2, np.sqrt(d2),
                                             notself=notself, permuted=True)
            if len(inds) == 0:
                return []
            I = np.argsort(D[:,0])
            inds = inds[I]
            rcfs = [[] for i in range(len(rds))]
            cols = T.columns()
            gotem = False
            if contains:
                if ('ramin' in cols and 'ramax' in cols and
                    'decmin' in cols and 'decmax' in cols):
                    gotem = True
                    for j,i in inds:
                        (r,d) = rds[i]
                        if (r >= T.ramin[j] and r <= T.ramax[j]
                            and d >= T.decmin[j] and d <= T.decmax[j]):
                            rcfs[i].append((T.run[j], T.camcol[j],
                                            T.field[j], T.ra[j], T.dec[j]))
                        #print '%i fields contain the first query RA,Dec' % len(rcfs[0])
                else:
                    print('you requested fields *containing* the query RA,Dec,')
                    print('but the fields list file \"%s\" doesn\'t contain RAMIN,RAMAX,DECMIN, and DECMAX columns' % tablefn)
            if not gotem:
                for j,i in inds:
                    rcfs[i].append((T.run[j], T.camcol[j], T.field[j],
                                    T.ra[j], T.dec[j]))

        if isscalar(ra) and isscalar(dec):
            return rcfs[0]
        return rcfs
        


# RA,Dec are either scalars or iterables.
# Radius is in *arcmin*.  sigh.
# If scalars, returns a list of (run, camcol, field, ra, dec) tuples, one for each matching field.
# If iterable, returns a list containing one list per query (ra,dec) of the same tuple.
def radec_to_sdss_rcf(ra, dec, spherematch=True, radius=0, tablefn=None, contains=False):
    RD = RaDecToRcf(tablefn=tablefn)
    return RD(ra, dec, spherematch=spherematch, radius=radius, contains=contains)



def OLD_radec_to_sdss_rcf(ra, dec, spherematch=True, radius=0, tablefn=None, contains=False):
    # This file is generated by merging the files "dr7_e.fits", "dr7_g.fits", and "dr7_a.fits",
    # whose construction is described in http://trac.astrometry.net/browser/trunk/projects/sdss-tests/README
    # (and in comments below that I didn't notice before writing this)
    if tablefn is None:
        tablefn = find_data_file('dr7fields.fits')
    sdss = table_fields(tablefn)
    if sdss is None:
        print('Failed to read table of SDSS fields from file', tablefn)
        raise Exception('Failed to read table of SDSS fields from file: "' + str(tablefn) + '"')
    sdssxyz = radectoxyz(sdss.ra, sdss.dec)
    ## HACK - magic 13x9 arcmin.
    if radius == 0:
        radius = sqrt(13.**2 + 9.**2)/2.
    radius2 = arcmin2distsq(radius)
    if not spherematch:
        rcfs = []
        for r,d in broadcast(ra,dec):
            xyz = radectoxyz(r,d)
            dist2s = sum((xyz - sdssxyz)**2, axis=1)
            I = flatnonzero(dist2s < radius2)
            if False:
                print('I:', I)
                print('fields:', sdss[I].run, sdss[I].field, sdss[I].camcol)
                print('RA', sdss[I].ra)
                print('Dec', sdss[I].dec)
            rcfs.append(zip(sdss[I].run, sdss[I].camcol, sdss[I].field, sdss[I].ra, sdss[I].dec))
    else:
        from astrometry.libkd import spherematch
        rds = array([x for x in broadcast(ra,dec)])
        xyz = radectoxyz(rds[:,0], rds[:,1]).astype(double)
        (inds,dists) = spherematch.match(xyz, sdssxyz, sqrt(radius2))
        #print 'found %i matches' % len(inds)
        if len(inds) == 0:
            return []
        #print 'inds:', inds.shape
        I = np.argsort(dists[:,0])
        #print 'dists:', dists.shape
        inds = inds[I,:]
        rcfs = [[] for i in range(len(rds))]
        cols = sdss.columns()
        gotem = False
        if contains:
            if 'ramin' in cols and 'ramax' in cols and 'decmin' in cols and 'decmax' in cols:
                gotem = True
                for i,j in inds:
                    (r,d) = rds[i]
                    if r >= sdss.ramin[j] and r <= sdss.ramax[j] and d >= sdss.decmin[j] and d <= sdss.decmax[j]:
                        rcfs[i].append((sdss.run[j], sdss.camcol[j], sdss.field[j], sdss.ra[j], sdss.dec[j]))
                print('%i fields contain the first query RA,Dec' % len(rcfs[0]))
            else:
                print('you requested fields *containing* the query RA,Dec,')
                print('but the fields list file \"%s\" doesn\'t contain RAMIN,RAMAX,DECMIN, and DECMAX columns' % tablefn)
        if not gotem:
            for i,j in inds:
                rcfs[i].append((sdss.run[j], sdss.camcol[j], sdss.field[j], sdss.ra[j], sdss.dec[j]))


    if isscalar(ra) and isscalar(dec):
        return rcfs[0]
    return rcfs

# The field list was created starting with dstn's list of fields in DR7:
#  fitscopy dr7_e.fits"[col RUN;FIELD;CAMCOL;RA=(RAMIN+RAMAX)/2;DEC=(DECMIN+DECMAX)/2]" e.fits
#  fitscopy dr7_g.fits"[col RUN;FIELD;CAMCOL;RA=(RAMIN+RAMAX)/2;DEC=(DECMIN+DECMAX)/2]" g.fits
#  fitscopy dr7_a.fits"[col RUN;FIELD;CAMCOL;RA=(RAMIN+RAMAX)/2;DEC=(DECMIN+DECMAX)/2]" a.fits
#  tabmerge g.fits e.fits
#  tabmerge g.fits+1 e.fits+1
#  tabmerge a.fits+1 e.fits+1
#  mv e.fits dr7fields.fits
#  rm g.fits a.fits

'''
cd ~/sdss-tests
casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS querywait @dr7_ngood.sql
casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS querywait @dr7_ngood2.sql
casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS outputdownloaddelete mydb.goodfields2 /tmp/dr7.fits
fitscopy /tmp/dr7.fits"[col RA=(ramin+ramax)/2;DEC=(decmin+decmax)/2;run;field;camcol;ngood;ramin;ramax;decmin;decmax]" dr7fields.fits

casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS querywait @s82_ngood.sql
# Stripe82 has no RunQA table.
casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS querywait @s82_ngood2.sql
casjobs.py $SDSS_CAS_USER $SDSS_CAS_PASS outputdownloaddelete mydb.s82goodfields2 s82.fits
fitscopy s82.fits"[col RA=(ramin+ramax)/2;DEC=(decmin+decmax)/2;run;field;camcol;ngood;ramin;ramax;decmin;decmax]" s82fields.fits

'''

def main():
    import sys
    from optparse import OptionParser
    
    parser = OptionParser(usage='%prog [options] <ra> <dec>')
    parser.add_option('-f', dest='fields', help='FITS table of fields to use; default is astrometry/data/dr7fields.fits')
    parser.add_option('-c', dest='contains', action='store_true', help='Print only fields that *contain* the given point; requires RAMIN,RAMAX,DECMIN,DECMAX fields.')
    parser.add_option('-b', '--bands', dest='bands', help='Retrieve fpCs of the given bands; default "ugriz"')
    parser.add_option('-t', dest='filetypes', help='Retrieve this file type (fpC, fpM, psField, tsField, tsObj, etc)', action='append', default=['fpC'])
    parser.add_option('-r', dest='radius', type=float, default=15., help='Search radius (arcmin)')
    parser.set_defaults(fields=None, contains=False, bands='ugriz')

    (opt, args) = parser.parse_args()
    if len(args) != 2:
        parser.print_help()
        print()
        print('Got extra arguments:', args)
        sys.exit(-1)

    # parse RA,Dec.
    try:
        ra = float(args[0])
    except ValueError:
        ra = hmsstring2ra(args[0])
    try:
        dec = float(args[1])
    except ValueError:
        dec = dmsstring2dec(args[1])
    
    tablefn = None
    if opt.fields is not None:
        if os.path.exists(opt.fields):
            tablefn = opt.fields
        else:
            tablefn = find_data_file(opt.fields)
        if tablefn is None:
            print('Failed to find list of fields:', opt.fields)
            sys.exit(-1)
    
    # arcmin
    radius = opt.radius
    rcfs = radec_to_sdss_rcf(ra,dec,radius=radius, tablefn=tablefn, contains=opt.contains)
    print('ra,dec', ra,dec)
    print('rcfs:', rcfs)
    print()
    for (r,c,f,ra1,dec1) in rcfs:
        print('%i %i %i (dist: %g arcmin)' % (r,c,f, deg2arcmin(degrees_between(ra,dec,ra1,dec1))))

    print()
    for (r,c,f,ra1,dec1) in rcfs:
        print('http://cas.sdss.org/dr7/en/get/frameByRCFZ.asp?R=%i&C=%i&F=%i&Z=0&submit1=Get+Image' % (r,c,f))

    print()
    for (r,c,f,ra1,dec1) in rcfs:
        print('wget "http://cas.sdss.org/dr7/en/get/frameByRCFZ.asp?R=%i&C=%i&F=%i&Z=0&submit1=Get+Image" -O sdss-%04i-%i-%04i.jpg' % (r,c,f,r,c,f))

    from .sdss_das import sdss_das_get
    for (r,c,f,ra1,dec1) in rcfs:
        for t in opt.filetypes:
            for b in opt.bands:
                R = sdss_das_get(t, None, r, c, f, b)
                if R is False:
                    continue
                if t == 'fpC':
                    fpc = sdss_filename('fpC', r, c, f, b)
                    os.system('gunzip -cd %s.gz > %s' % (fpc,fpc))
                    wcs = Tan(filename=fpc)
                    x,y = wcs.radec2pixelxy(ra, dec)
                    x,y = int(x),int(y)
                    os.system('imcopy %s"[%i:%i,%i:%i]" !/tmp/cut-%s' % (fpc, max(0, x-100), x+100, max(0, y-100), y+100, fpc))
                    os.system('an-fitstopnm -i /tmp/cut-%s -N 1150 -X 1400 | pnmtopng > cut-%s.png' % (fpc, fpc))
                    print('R,C,F', r,c,f)
                    print('x,y', x,y)

if __name__ == '__main__':
    main()




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from .dr7 import DR7
from .dr8 import DR8
from .dr9 import DR9
from .dr10 import DR10

from .common import band_name, band_index, band_names, cas_flags
from .common import photo_flags1_info, photo_flags2_info, photo_flags1_map, photo_flags2_map
from .common import munu_to_radec_deg, AsTransWrapper, AsTrans

from .fields import *



#! /usr/bin/env python3
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

import sys
from optparse import OptionParser

import matplotlib
matplotlib.use('Agg')

from pylab import *
from numpy import *

from astrometry.libkd import spherematch

def plotshift(ixy, rxy, dcell=50, ncells=18, outfn=None, W=None, H=None, hist=False,
              nhistbins=21):
              #histbinsize=None):
    # correspondences we could have hit...
    radius = dcell * sqrt(2.)
    #print 'ixy', ixy.shape
    #print 'rxy', rxy.shape
    assert((len(rxy) == 0) or (rxy.shape[1] == 2))
    assert((len(ixy) == 0) or (ixy.shape[1] == 2))

    ix = ixy[:,0]
    iy = ixy[:,1]

    if W is None:
        W = max(ix)
    if H is None:
        H = max(iy)

    if len(rxy):
        keep = (rxy[:,0] > -dcell) * (rxy[:,0] < W+dcell) * (rxy[:,1] > -dcell) * (rxy[:,1] < H+dcell)
        rxy = rxy[keep]
    #print 'Cut to %i ref sources in range' % len(rxy)
    
    cellsize = sqrt(W * H / ncells)
    nw = int(round(W / cellsize))
    nh = int(round(H / cellsize))
    #print 'Grid cell size', cellsize
    #print 'N cells', nw, 'x', nh
    edgesx = linspace(0, W, nw+1)
    edgesy = linspace(0, H, nh+1)
    #print 'Edges:'
    #print '     x:', edgesx
    #print '     y:', edgesy
    if len(rxy) == 0:
        binx = array([])
        biny = array([])
    else:
        binx = digitize(rxy[:,0], edgesx)
        biny = digitize(rxy[:,1], edgesy)
    binx = clip(binx - 1, 0, nw-1)
    biny = clip(biny - 1, 0, nh-1)
    bin = biny * nw + binx

    clf()

    for i in range(nh):
        for j in range(nw):
            thisbin = i * nw + j
            R = (bin == thisbin)
            #print 'cell %i, %i' % (j, i)
            #print '%i ref sources' % sum(R)
            matchdx = []

            if sum(R) > 0:
                (inds,dists) = spherematch.match(rxy[R,:], ixy, radius)
                #print 'Found %i matches within %g pixels' % (len(dists), radius)
                ri = inds[:,0]
                # un-cut ref inds...
                ri = (flatnonzero(R))[ri]
                ii = inds[:,1]
                matchx    = rxy[ri,0]
                matchy    = rxy[ri,1]
                matchdx = ix[ii] - matchx
                matchdy = iy[ii] - matchy
                #print 'All matches:'
                #for dx,dy in zip(matchdx,matchdy):
                #    print '     %.1f, %.1f' % (dx,dy)
                ok = (matchdx >= -dcell) * (matchdx <= dcell) * (matchdy >= -dcell) * (matchdy <= dcell)
                matchdx = matchdx[ok]
                matchdy = matchdy[ok]
                #print 'Cut to %i within %g x %g square' % (sum(ok), dcell*2, dcell*2)
                #print 'Cut matches:'
                #for dx,dy in zip(matchdx,matchdy):
                #    print '     %.1f, %.1f' % (dx,dy)
            
            # Subplot places plots left-to-right, TOP-to-BOTTOM.
            subplot(nh, nw, 1 + ((nh - i - 1)*nw + j))

            if len(matchdx) > 0:
                #plot(matchdx, matchdy, 'ro', mec='r', mfc='r', ms=5, alpha=0.2)
                #plot(matchdx, matchdy, 'ro', mec='r', mfc='none', ms=5, alpha=0.2)
                if hist:
                    #if histbinsize is None:
                    #    histbinsize = dcell / 10.
                    edges = linspace(-dcell, dcell, nhistbins+1)
                    (H,xe,ye) = histogram2d(matchdx, matchdy, bins=(edges,edges))
                    imshow(H.T, extent=(min(xe), max(xe), min(ye), max(ye)),
                           aspect='auto', origin='lower', interpolation='nearest')
                    text(dcell, dcell, '%i' % H.max(), color='y',
                         horizontalalignment='right', verticalalignment='top')
                         
                else:
                    plot(matchdx, matchdy, 'r.', alpha=0.3)

            if hist:
                axhline(0, color='b', alpha=0.8)
                axvline(0, color='b', alpha=0.8)
            else:
                axhline(0, color='k', alpha=0.5)
                axvline(0, color='k', alpha=0.5)
            if i == 0 and j == 0:
                xticks([-dcell,0,dcell])
                yticks([-dcell,0,dcell])
            else:
                xticks([],[])
                yticks([],[])
            axis('scaled')
            axis([-dcell, dcell, -dcell, dcell])
    if outfn is not None:
        #print 'Saving', outfn
        savefig(outfn)


if __name__ == '__main__':
    from astrometry.util.fits import fits_table
    
    parser = OptionParser('usage: %prog [options] <image xy> <reference xy> <plot name>')
    parser.add_option('-X', dest='xcol', help='Name of X column in image table')
    parser.add_option('-Y', dest='ycol', help='Name of Y column in image table')
    parser.add_option('-N', dest='nimage', type='int', help='Cut to the first N image sources')
    parser.add_option('-x', dest='rxcol', help='Name of X column in reference table')
    parser.add_option('-y', dest='rycol', help='Name of Y column in reference table')
    parser.add_option('-n', dest='nref', type='int', help='Cut to the first N reference sources')
    parser.add_option('-c', dest='cells', type='int', help='Approx. number of pieces to cut image into (default:18)')
    parser.add_option('-s', dest='cellsize', type='int', help='Search radius, in pixels (default 50)')
    parser.set_defaults(xcol='X', ycol='Y', nimage=0, cells=0, cellsize=0, rxcol='X', rycol='Y', nref=0)
    opt,args = parser.parse_args()

    if len(args) != 3:
        parser.print_help()
        sys.exit(-1)

    imxy = fits_table(args[0])
    refxy = fits_table(args[1])
    outfn = args[2]

    kwargs = {}
    if opt.cells:
        kwargs['ncells'] = opt.cells
    if opt.cellsize:
        kwargs['dcell'] = opt.cellsize

    ix = imxy.getcolumn(opt.xcol)
    iy = imxy.getcolumn(opt.ycol)
    ixy = vstack((ix,iy)).T
    if opt.nimage:
        ixy = ixy[:opt.nimage,:]

    rx = refxy.getcolumn(opt.rxcol)
    ry = refxy.getcolumn(opt.rycol)
    rxy = vstack((rx,ry)).T
    if opt.nref:
        rxy = rxy[:opt.nref,:]

    plotshift(ixy, rxy, outfn=outfn, **kwargs)



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
import os

from .common import *

from astrometry.util.miscutils import *
from astrometry.util.fits import *

class DR7(SdssDR):
    def __init__(self, **kwargs):
        '''
        kwargs:
        
        (base class:)
        curl=False: use curl rather than wget?
        basedir=None: base directory for local files
        '''
        SdssDR.__init__(self, **kwargs)
        # These are *LOCAL* filenames -- some are different than those
        # on the DAS.
        self.filenames = {
            'fpObjc': 'fpObjc-%(run)06i-%(camcol)i-%(field)04i.fit',
            'fpM': 'fpM-%(run)06i-%(band)s%(camcol)i-%(field)04i.fit',
            'fpC': 'fpC-%(run)06i-%(band)s%(camcol)i-%(field)04i.fit',
            'fpAtlas': 'fpAtlas-%(run)06i-%(camcol)i-%(field)04i.fit',
            'psField': 'psField-%(run)06i-%(camcol)i-%(field)04i.fit',
            #'tsObj': 'tsObj-%(run)06i-%(camcol)i-%(rerun)i-%(field)04i.fit',
            #'tsField': 'tsField-%(run)06i-%(camcol)i-%(rerun)i-%(field)04i.fit',
            'tsObj': 'tsObj-%(run)06i-%(camcol)i-%(field)04i.fit',
            'tsField': 'tsField-%(run)06i-%(camcol)i-%(field)04i.fit',
            }
        self.softbias = 1000

    def getDRNumber(self):
        return 7
        
    def retrieve(self, filetype, run, camcol, field, band=None, skipExisting=True):
        # FIXME!
        from astrometry.util.sdss_das import sdss_das_get
        outfn = self.getPath(filetype, run, camcol, field, band)
        #print 'Output filename:', outfn
        if skipExisting and os.path.exists(outfn):
            return
        return sdss_das_get(filetype, outfn, run, camcol, field, band,
                            curl=self.curl)

    def readTsField(self, run, camcol, field, rerun):
        '''
        http://www.sdss.org/dr7/dm/flatFiles/tsField.html

        band: string ('u', 'g', 'r', 'i', 'z')
        '''
        f = TsField(run, camcol, field, rerun=rerun)
        fn = self.getFilename('tsField', run, camcol, field, rerun=rerun)
        #print 'reading file', fn
        p = self._open(fn)
        #print 'got', len(p), 'HDUs'
        f.setHdus(p)
        return f

    def readFpC(self, run, camcol, field, band):
        '''
        http://data.sdss3.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/objcs/CAMCOL/fpC.html

        band: string ('u', 'g', 'r', 'i', 'z')
        '''
        f = FpC(run, camcol, field, band)
        # ...
        fn = self.getFilename('fpC', run, camcol, field, band)
        #print 'reading file', fn
        p = self._open(fn)
        #print 'got', len(p), 'HDUs'
        f.image = p[0].data
        f.header = p[0].header
        return f

    def readFpObjc(self, run, camcol, field):
        '''
        http://data.sdss3.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/objcs/CAMCOL/fpObjc.html
        '''
        f = FpObjc(run, camcol, field)
        # ...
        fn = self.getFilename('fpObjc', run, camcol, field)
        #print 'reading file', fn
        p = self._open(fn)
        #print 'got', len(p), 'HDUs'
        return f

    def readFpM(self, run, camcol, field, band):
        '''
        http://data.sdss3.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/objcs/CAMCOL/fpM.html
        '''
        f = FpM(run, camcol, field, band)
        # ...
        fn = self.getFilename('fpM', run, camcol, field, band)
        #print 'reading file', fn
        p = self._open(fn)
        #print 'got', len(p), 'HDUs'
        f.setHdus(p)
        return f

    def readPsField(self, run, camcol, field):
        '''
        http://data.sdss3.org/datamodel/files/PHOTO_REDUX/RERUN/RUN/objcs/CAMCOL/psField.html
        '''
        f = PsField(run, camcol, field)
        # ...
        fn = self.getFilename('psField', run, camcol, field)
        #print 'reading file', fn
        p = self._open(fn)
        #print 'got', len(p), 'HDUs'
        f.setHdus(p)
        return f

    def getInvvar(self, fpC, fpM, gain, darkvar, sky, skyerr,
                  x0=0, x1=None, y0=0, y1=None, invvar_and_mask=False):
        '''
        Produces a (cut-out of) the inverse-variance noise image, from columns
        [x0,x1] and rows [y0,y1] (inclusive).  Default is the whole image.

        fpC is the image pixels (eg FpC.getImage())
        #### CHECK THIS -- below we have  (img + sky), but fpCs have *not*
        had sky subtracted.

        fpM is the FpM
        gain, darkvar, sky, and skyerr can be retrieved from the psField file.
        '''
        if x1 is None:
            x1 = fpC.shape[1]-1
        if y1 is None:
            y1 = fpC.shape[0]-1

        # Poisson: mean = variance
        # Add readout noise?
        # Spatial smoothing?
        img = fpC[y0:y1+1, x0:x1+1]

        # from http://www.sdss.org/dr7/algorithms/fluxcal.html
        ivarimg = 1./((img + sky) / gain + darkvar + skyerr)

        if invvar_and_mask:
            mask = np.ones(ivarimg.shape, np.bool)
            maskimg = mask
        else:
            maskimg = ivarimg
        
        # Noise model:
        #  -mask coordinates are wrt fpC coordinates.
        #  -INTERP, SATUR, CR,
        #  -GHOST?
        for plane in [ 'INTERP', 'SATUR', 'CR', 'GHOST' ]:
            fpM.setMaskedPixels(plane, maskimg, 0)

        if invvar_and_mask:
            return ivarimg, mask
        return ivarimg





#! /usr/bin/env python3
"""
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE

Convert an image in a variety of formats into a pnm file
"""
from __future__ import print_function
from __future__ import absolute_import
import sys
import os
import os.path
import tempfile

from astrometry.util.shell import shell_escape
from astrometry.util.filetype import filetype_short
import logging

fitstype = 'FITS image data'
fitsext = 'fits'
tiffext = 'tiff'

pgmcmd = 'pgmtoppm rgbi:1/1/1 %s > %s'
pgmext = 'pgm'

an_fitstopnm_ext_cmd = 'an-fitstopnm -e %i -i %%s > %%s'

imgcmds = {fitstype : (fitsext, 'an-fitstopnm -i %s > %s'),
       'JPEG image data'  : ('jpg',  'jpegtopnm %s > %s'),
       'PNG image data'       : ('png',      'pngtopnm %s > %s'),
       'PNG image'    : ('png',      'pngtopnm %s > %s'),
       'GIF image data'       : ('gif',      'giftopnm %s > %s'),
       'Netpbm PPM'       : ('ppm',      'ppmtoppm < %s > %s'),
       'Netpbm PPM "rawbits" image data' : ('ppm',  'cp %s %s'),
       'Netpbm PGM'       : ('pgm',      pgmcmd),
       'Netpbm PGM "rawbits" image data' : ('pgm',  pgmcmd),
       'TIFF image data'  : ('tiff',  'tifftopnm %s > %s'),
       'PC bitmap' : ('bmp', 'bmptopnm %s > %s'),
       # RAW is not recognized by 'file'; we have to use 'dcraw',
       # but we still store this here for convenience.
       'raw'              : ('raw', 'dcraw -4 -c %s > %s'),
       }

compcmds = {'gzip compressed data'    : ('gz',      'gunzip -c %s > %s'),
        "compress'd data 16 bits" : ('gz',      'gunzip -c %s > %s'),
        'bzip2 compressed data'   : ('bz2', 'bunzip2 -k -c %s > %s')
        }

funpack_cmd = 'funpack -E %i -S %s > %s'

# command to identify a RAW image.
raw_id_cmd = 'dcraw -i %s >/dev/null 2> /dev/null'

verbose = False

def do_command(cmd):
    logging.debug('Running: "%s"' % cmd)
    if os.system(cmd) != 0:
        print('Command failed: %s' % cmd, file=sys.stderr)
        sys.exit(-1)

def get_cmd(types, cmds):
    if types is None:
        return None
    ext=None
    cmd=None
    for t in types:
        (ext,cmd) = cmds.get(t, (None,None))
        if ext is not None:
            break
    return (ext,cmd)

def uncompress_file(infile, uncompressed, typeinfo=None, extension=None):
    """
    infile: input filename.
    uncompressed: output filename.
    typeinfo: output from the 'file' command; if None we'll run 'file'.

    Returns: comptype
    comptype: None if the file wasn't compressed, or 'gz' or 'bz2'.
    """
    if typeinfo is None:
        typeinfo = filetype_short(infile)
        if typeinfo is None:
            logging.debug('Could not determine file type of "%s"' % infile)
            return None
    # print('uncompress_file: type is', typeinfo)
    (ext,cmd) = get_cmd(typeinfo, compcmds)
    # print('ext:', ext)
    if ext is None:
        # Check for fpack compressed FITS file.
        if fitstype in typeinfo:
            # FITS file.  Check header for ZIMAGE=T
            try:
                import fitsio
                logging.debug('Checking FITS header of', infile, 'ext',
                              extension, 'for ZIMAGE card (fpack compression)')
                hdr = fitsio.read_header(infile, ext=extension)
                if hdr.get('ZIMAGE', False):
                    # Compressed
                    cmd = (funpack_cmd % (
                        extension or 0,
                        shell_escape(infile), shell_escape(uncompressed)))
                    logging.debug('Fpack compressed; uncompressing with', cmd)
                    if os.system(cmd) == 0:
                        return 'fz'
            except:
                pass
        logging.debug('File is not compressed: "%s"' % '/'.join(typeinfo))
        return None
    assert uncompressed != infile
    logging.debug('Compressed file (type %s), dumping to: "%s"' % (ext, uncompressed))
    do_command(cmd % (shell_escape(infile), shell_escape(uncompressed)))
    return ext

def is_raw(fn):
    rtn = os.system(raw_id_cmd % shell_escape(fn))
    logging.debug('ran dcraw: return value %i' % rtn)
    return os.WIFEXITED(rtn) and (os.WEXITSTATUS(rtn) == 0)

# Returns (extension, command, error)
def get_image_type(infile):
    typeinfo = filetype_short(infile)
    if typeinfo is None:
        return (None, None, 'Could not determine file type (does the file exist?): %s' % infile)
    (ext,cmd) = get_cmd(typeinfo, imgcmds)
    logging.debug('ext: %s' % ext)
    # "file" recognizes some RAWs as TIFF, but tifftopnm can't read them...
    # run "dcraw" here if the type is TIFF.
    if ext == tiffext and is_raw(infile):
        (ext, cmd) = imgcmds['raw']
    if ext is not None:
        return (ext, cmd, None)
    if ext != tiffext and is_raw(infile):
        # it's a RAW image.
        (ext, cmd) = imgcmds['raw']
        return (ext, cmd, None)
    return (None, None, 'Unknown image type "%s"' % typeinfo)

def find_program(dirs, cmd):
    # pull off the executable name.
    parts = cmd.split(' ', 1)
    prog = parts[0]
    # try the same directory - this should work for installed
    # versions where image2pnm.py and an-fitstopnm are both in
    # "bin".
    for mydir in dirs:
        # If mydir is actually a file, get its dir
        if os.path.isfile(mydir):
            mydir = os.path.dirname(mydir)
        p = os.path.join(mydir, prog)
        if os.path.exists(p):
            return ' '.join([p, parts[1]])
        logging.info('path', p, 'does not exist.')
    return None

def image2pnm(infile, outfile, force_ppm=False, extension=None, mydir=None):
    """
    infile: input filename.
    outfile: output filename.
    force_ppm: boolean, convert PGM to PPM so that the output is always PPM.

    Returns: (type, error)

    - type: (string): image type: 'jpg', 'png', 'gif', etc., or None if
       image type isn't recognized.

    - error: (string): error string, or None
    """
    (ext, cmd, err) = get_image_type(infile)
    if ext is None:
        return (None, 'Image type not recognized: ' + err)

    tempfiles = []
    (outfile_dir, outfile_file) = os.path.split(outfile)

    if force_ppm:
        original_outfile = outfile
        outfile_dir = os.path.dirname(outfile)
        (f, outfile) = tempfile.mkstemp(suffix='.pnm',
                    dir=outfile_dir)
        # we might rename this file later, so don't add it to the list of
        # tempfiles to delete until later...
        os.close(f)
        logging.debug('temporary output file: %s' % outfile)

    if ext == fitsext and extension:
        cmd = an_fitstopnm_ext_cmd % extension

    if ext == fitsext and mydir:
        # an-fitstopnm: add explicit path...
        cmd = find_program(mydir, cmd)
        if cmd is None:
            return (None, 'Couldn\'t find the program "an-fitstopnm".')

    # Do the actual conversion
    do_command(cmd % (shell_escape(infile), shell_escape(outfile)))

    if force_ppm:
        if ext == pgmext:
            # Convert to PPM.
            do_command(pgmcmd % (shell_escape(outfile), shell_escape(original_outfile)))
            tempfiles.append(outfile)
        else:
            # print 'file type extension:', ext, '; renaming', outfile, 'to', original_outfile
            os.rename(outfile, original_outfile)

    for fn in tempfiles:
        os.unlink(fn)
    # Success
    return (ext, None)

def convert_image(infile, outfile, uncompressed=None, force_ppm=False,
                  extension=None, mydir=None):
    tempfiles = []
    # if the caller didn't specify where to put the uncompressed file,
    # create a tempfile.
    if uncompressed is None:
        (outfile_dir, outfile_file) = os.path.split(outfile)
        (f, uncompressed) = tempfile.mkstemp('', 'uncomp', outfile_dir)
        os.close(f)
        tempfiles.append(uncompressed)
    comp = uncompress_file(infile, uncompressed, extension=extension)
    if comp:
        print('compressed')
        print(comp)
        infile = uncompressed
        if comp == 'fz':
            # Funpack writes a single-HDU output file.
            extension = 0

    (imgtype, errstr) = image2pnm(infile, outfile, force_ppm=force_ppm,
                                  extension=extension, mydir=mydir)
    for fn in tempfiles:
        os.unlink(fn)

    if errstr:
        logging.error('ERROR: %s' % errstr)
        return -1
    print(imgtype)
    return 0

def main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('-i', '--infile',
                      dest='infile',
                      help='input image FILE', metavar='FILE')
    parser.add_option('-u', '--uncompressed-outfile',
                      dest='uncompressed_outfile',
                      help='uncompressed temporary FILE', metavar='FILE',
                      default='')
    parser.add_option('-o', '--outfile',
                      dest='outfile',
                      help='output pnm image FILE', metavar='FILE')
    parser.add_option('-p', '--ppm',
                      action='store_true', dest='force_ppm',
                      help='convert the output to PPM')
    parser.add_option('-e', '--extension',
                      dest='extension', type='int',
                      help='FITS extension to read')
    parser.add_option('--mydir', help='Set directory to search for an-fitstopnm')
    parser.add_option('-v', '--verbose',
                      action='store_true', dest='verbose',
                      help='be chatty')

    (options, args) = parser.parse_args()

    if not options.infile:
        parser.error('required argument missing: infile')
    if not options.outfile:
        parser.error('required argument missing: outfile')

    # Find the path to this executable and use it to find other Astrometry.net
    # executables.
    dirs = []
    if options.mydir:
        dirs.append(options.mydir)
    if len(sys.argv):
        dirs.append(os.path.dirname(sys.argv[0]))
    # util/ -- useful when running from source directory
    dirs.append(os.path.dirname(__file__))
    print('Dirs:', dirs)
    
    global verbose
    verbose = options.verbose

    logformat = '%(message)s'
    if verbose:
        logging.basicConfig(level=logging.DEBUG, format=logformat)
    else:
        logging.basicConfig(level=logging.INFO, format=logformat)
    logging.raiseExceptions = False
        
    return convert_image(options.infile, options.outfile,
                         uncompressed=options.uncompressed_outfile,
                         force_ppm=options.force_ppm,
                         extension=options.extension,
                         mydir=dirs)

if __name__ == '__main__':
    sys.exit(main())



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from distutils.command.build_ext import build_ext # as _build_py
#from distutils.command import build_ext as _build_ext
from distutils.dep_util import newer_group, newer
from distutils import log
import os

class an_build_ext(build_ext):
    '''SWIG dependency checking.'''
    def swig_sources(self, sources, extension):
        ### Copied from build_ext.py : swig_sources, with small mods marked "# AN"
        new_sources = []
        swig_sources = []
        swig_targets = {}
        # AN
        swig_py_targets = {}
        if self.swig_cpp:
            log.warn("--swig-cpp is deprecated - use --swig-opts=-c++")

        if self.swig_cpp or ('-c++' in self.swig_opts) or \
          ('-c++' in extension.swig_opts):
            target_ext = '.cpp'
        else:
            target_ext = '.c'

        for source in sources:
            (base, ext) = os.path.splitext(source)
            if ext == ".i":                # SWIG interface file
                new_sources.append(base + '_wrap' + target_ext)
                swig_sources.append(source)
                swig_targets[source] = new_sources[-1]
                # AN
                swig_py_targets[source] = base + '.py'
            else:
                new_sources.append(source)

        if not swig_sources:
            return new_sources

        swig = self.swig or self.find_swig()
        swig_cmd = [swig, "-python"]
        swig_cmd.extend(self.swig_opts)
        if self.swig_cpp:
            swig_cmd.append("-c++")

        # Do not override commandline arguments
        if not self.swig_opts:
            for o in extension.swig_opts:
                swig_cmd.append(o)

        for source in swig_sources:
            target = swig_targets[source]
            log.info("swigging %s to %s", source, target)
            # AN
            py_target = swig_py_targets[source]
            if not (self.force or
                    newer(source, target) or newer(source, py_target)):
                log.debug("skipping swig of %s (older than %s, %s)" %
                          (source, target, py_target))
                continue
            # AN
            
            self.spawn(swig_cmd + ["-o", target, source])

        return new_sources



# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
from __future__ import absolute_import

import os
import re
import resource

# split out of get_memusage for profiling purposes
def _read_proc_status(pid):
    procfn = '/proc/%d/status' % pid
    t = open(procfn).readlines()
    d = dict([(line.split()[0][:-1], line.split()[1:]) for line in t])
    return d

                 
def _read_proc_maps(pid):
    procfn = '/proc/%d/maps' % pid
    t = open(procfn).readlines()
    d = dict(mmaps=t)
    rex = re.compile(r'(?P<addrlo>[0-9a-f]+)-(?P<addrhi>[0-9a-f]+) .*')
    parsed = []
    addrsum = 0
    for line in t:
        m = rex.match(line)
        if m is not None:
            parsed.append(m.groupdict())
            try:
                addrsum += int(m.group('addrhi'), 16) - int(m.group('addrlo'), 16)
            except:
                pass
    return dict(mmaps=t, mmaps_parsed=parsed, mmaps_total=addrsum)

def get_memusage(mmaps=True):
    '''
    Retrieves memory usage information from the *resource* module and from
    reading the */proc* filesystem.

    Returns:
    --------
    Dictionary of *quantity* to *(value, unit)* tuples.  Value is a float,
    unit is "MB", for example.
    '''
    ru = resource.getrusage(resource.RUSAGE_SELF)
    #pgsize = resource.getpagesize()
    maxrss = float(ru.ru_maxrss) / 1e6
    # print('shared memory size:', (ru.ru_ixrss / 1e6), 'MB')
    # print('unshared memory size:', (ru.ru_idrss / 1e6), 'MB')
    # print('unshared stack size:', (ru.ru_isrss / 1e6), 'MB')
    # print('shared memory size:', ru.ru_ixrss)
    # print('unshared memory size:', ru.ru_idrss)
    # print('unshared stack size:', ru.ru_isrss)
    mu = dict(maxrss=[maxrss, 'MB'])

    pid = os.getpid()
    try:
        mu.update(_read_proc_status(pid))
    except:
        pass

    # /proc/meminfo ?

    if mmaps:
        try:
            mu.update(_read_proc_maps(pid))
        except:
            pass
    
    return mu

def get_procio():
    procfn = '/proc/%d/io' % os.getpid()
    dd = dict()
    try:
        t = open(procfn).readlines()
        for line in t:
            words = line.split()
            key = words[0].strip().strip(':')
            val = int(words[1])
            dd[key] = val
    except:
        pass
    return dd
    
def memusage():
    '''
    Prints current memory usage.
    '''
    mu = get_memusage()
    print('Memory usage:')
    print('max rss:', mu['maxrss'], 'MB')
    for key in ['VmPeak', 'VmSize', 'VmRSS', 'VmData', 'VmStk'
                # VmLck, VmHWM, VmExe, VmLib, VmPTE
                ]:
        print(key, ' '.join(mu.get(key, [])))
    if 'mmaps' in mu:
        print('Number of mmaps:', len(mu['mmaps']))

def count_file_descriptors():
    '''
    Returns the number of file descriptors used by the current process.
    On Linux this uses the */proc* filesystem; on Mac OSX it uses the *lsof*
    command.
    '''
    procfn = '/proc/%d/fd' % os.getpid()
    try:
        # Linux: /proc/PID/fd/*
        t = os.listdir(procfn)
        return len(t)
    except:
        pass

    try:
        # OSX: "lsof"
        from .run_command import run_command as rc
        cmd = 'lsof -p %i' % os.getpid()
        rtn,out,err = rc(cmd)
        if rtn == 0:
            return len(out.split('\n'))
    except:
        pass
    return 0

class FileDescriptorMeas(object):
    '''
    A measurement class for the *Time* class that measures file descriptors.
    '''
    def __init__(self):
        self.fds = count_file_descriptors()
    def format_diff(self, other):
        return 'Open files: %i' % self.fds
            
class MemMeas(object):
    '''
    A measurement class for the *Time* class that measures memory usage.
    '''
    def __init__(self):
        self.mem0 = get_memusage(mmaps=False)
    def format_diff(self, other):
        txt = []
        for k in ['VmPeak', 'VmSize', 'VmRSS', 'VmData', 'maxrss']:
            if not k in self.mem0:
                continue
            val,unit = self.mem0[k]
            if unit == 'kB':
                val = int(val, 10)
                val /= 1024.
                unit = 'MB'
                val = '%.0f' % val
            txt.append('%s: %s %s' % (k, val, unit))
        return ', '.join([] + txt)

class IoMeas(object):
    '''
    A measurement class for the *Time* class that measures disk I/O.
    '''    
    def __init__(self):
        self.io0 = get_procio()
    def format_diff(self, other):
        txt = []
        d1 = self.io0
        d0 = other.io0
        for k,knice in [('rchar',None), ('wchar',None),
                        ('read_bytes','rb'), ('write_bytes', 'wb')]:
            v1 = d1.get(k)
            v0 = d0.get(k)
            unit = 'b'
            dv = float(v1 - v0)
            for uu in ['kB', 'MB', 'GB']:
                if dv < 2048:
                    break
                dv = dv / 1024.
                unit = uu
            val = '%.3g' % dv
            if knice is None:
                kk = k
            else:
                kk = knice
            txt.append('%s: %s %s' % (kk, val, unit))
        return ', '.join([] + txt)

class CpuMeas(object):
    '''
    A measurement class for the *Time* class that measures CPU and Wall tim.
    '''    
    def __init__(self):
        import datetime
        from time import process_time
        self.wall = datetime.datetime.now()
        self.cpu = process_time()

    def cpu_seconds_since(self, other):
        '''
        Returns the CPU time used since the *other* timestamp was taken.

        eg::

            t0 = CpuMeas()
            # do stuff
            t1 = CpuMeas()
            dt = t1.cpu_seconds_since(t0)
            print('That took', dt, 'seconds of CPU time')
        '''
        return self.cpu - other.cpu

    def wall_seconds_since(self, other):
        '''
        Returns the wall-clock time in seconds since the *other* timestamp
        was taken.

        eg::
        
            t0 = CpuMeas()
            # do stuff
            t1 = CpuMeas()
            dt = t1.wall_seconds_since(t0)
            print('That took', dt, 'seconds')
        '''
        dwall = (self.wall - other.wall)
        # python2.7
        if hasattr(dwall, 'total_seconds'):
            dwall = dwall.total_seconds()
        else:
            dwall = (dwall.microseconds + (dwall.seconds + dwall.days * 24. * 3600.) * 1e6) / 1e6
        return dwall
        
    def format_diff(self, other):
        dwall = self.wall_seconds_since(other)
        dcpu = self.cpu_seconds_since(other)
        return 'Wall: %.2f s, CPU: %.2f s' % (dwall, dcpu)
        
class Time(object):
    '''
    A class for recording how much time (or other resources) are used
    by a process.  The *CpuMeas* class is used by default; others can
    be added as desired.

    Use like this::

        from astrometry.util.ttime import Time, MemMeas

        # measure memory usage too
        Time.add_measurement(MemMeas)
        
        def mymethod():
            t0 = Time()
            # do stuff
            t1 = Time()
            print('Time taken:', t1-t0)
    '''
    @staticmethod
    def add_measurement(m):
        Time.measurements.append(m)
    @staticmethod
    def remove_measurement(m):
        Time.measurements.remove(m)
    measurements = [CpuMeas]

    def __init__(self):
        self.meas = [m() for m in Time.measurements]

    def __sub__(self, other):
        '''
        Returns a string representation of the difference: self - other
        '''
        meas = ', '.join([m.format_diff(om) for m,om in zip(self.meas, other.meas)])
        return meas




# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
from __future__ import print_function
import matplotlib
if __name__ == '__main__':
    matplotlib.use('Agg')
import unittest
try:
    import pyfits
except ImportError:
    try:
        from astropy.io import fits as pyfits
    except ImportError:
        raise ImportError("Cannot import either pyfits or astropy.io.fits")
from astrometry.util.fits import *
from astrometry.plot.plotstuff import *
import numpy as np
import pylab as plt
from math import pi,sqrt

class TestPlotstuff2(unittest.TestCase):

    def testSubpixelShift(self):
        W,H = 100,100
        cx,cy = 34.2, 56.7
        X,Y = np.meshgrid(np.arange(W), np.arange(H))
        S = 2.
        G = 1./(2.*pi*S**2) * np.exp(-((X-cx)**2+(Y-cy)**2)/(2.*S**2))
        #G /= G.max()
        print(G.sum())
        print(np.sum(G*X), np.sum(G*Y))
        imfn = 'test-plotstuff-2.fits'
        pyfits.writeto(imfn, G, clobber=True)

        wcs = anwcs_create_box(33.5, 10.4, 1., W, H)
        anwcs_rotate_wcs(wcs, 25.)
        wcsfn = 'test-plotstuff-2.wcs'
        anwcs_write(wcs, wcsfn)

        plot = Plotstuff()
        plot.outformat = PLOTSTUFF_FORMAT_PNG
        plot.size = (W,H)
        plot.wcs_file = wcsfn
        plot.color = 'black'
        plot.plot('fill')

        im = plot.image
        im.image_low = 0
        im.image_high = G.max()
        plot_image_set_wcs(im, wcsfn, 0)
        plot_image_set_filename(im, imfn)
        plot.plot('image')

        plotfn = 'test-plotstuff-2.png'
        plot.write(plotfn)

        I = plt.imread(plotfn)
        I = I[:,:,0]
        sx,sy = (I*X).sum()/I.sum(), (I*Y).sum()/I.sum()
        print(sx,sy)
        ex,ey = cx,cy
        self.assertTrue(abs(sx-ex) < 0.1)
        self.assertTrue(abs(sy-ey)< 0.1)

        # Shift the plot's WCS CRPIX.
        dx = 0.25
        dy = 0.3
        sip = anwcs_get_sip(wcs)
        tan = sip.wcstan
        plotstuff_set_wcs(plot.pargs, wcs)

        xy = plot.xy
        plot_xy_set_wcs_filename(xy, wcsfn)
        plot.marker = 'crosshair'

        # Move the plot WCS origin
        for step in range(16):
            if step < 8:
                tan.set_crpix(tan.crpix[0]+dx, tan.crpix[1])
                ex += dx
            else:
                tan.set_crpix(tan.crpix[0], tan.crpix[1]+dy)
                ey += dy
            #anwcs_print_stdout(wcs)
            plot.color = 'black'
            plot.plot('fill')
            plot.plot('image')
            plotfn = 'test-plotstuff-2b%i.png' % step
            plot.write(plotfn)
            #anwcs_write(plot.pargs.wcs, 'test-plotstuff-2b.wcs')
            I = plt.imread(plotfn)
            I = I[:,:,0]
            sx,sy = (I*X).sum()/I.sum(), (I*Y).sum()/I.sum()
            print(sx,sy)
            print(ex,ey)
            self.assertTrue(abs(sx-ex) < 0.1)
            self.assertTrue(abs(sy-ey) < 0.1)

            plot.color = 'red'
            plot_xy_clear_list(xy)
            # don't plot at sx,sy / ex,ey -- original image coords
            # are unchanged.
            plot_xy_vals(xy, cx+1, cy+1)
            plot.plot('xy')
            plotfn = 'test-plotstuff-2c%i.png' % step
            plot.write(plotfn)
            # visual check that plot xy symbols match source centroid -- yes
            

        # Scan image WCS in RA,Dec; check that recovered source position
        # through plot WCS matches.
        # reset...
        plot.wcs_file = wcsfn
        plotwcs = anwcs_open(wcsfn, 0)

        wcs = anwcs_open(wcsfn, 0)
        im.wcs = wcs
        sip = anwcs_get_sip(wcs)
        tan = sip.wcstan
        ddec = 1.2 * 1./W
        ok,era,edec = anwcs_pixelxy2radec(wcs, cx, cy)
        print(era,edec)
        for step in range(16):
            tan.set_crval(tan.crval[0], tan.crval[1] + ddec)
            edec += ddec

            plot.color = 'black'
            plot.plot('fill')
            plot.plot('image')
            plotfn = 'test-plotstuff-2d%i.png' % step
            plot.write(plotfn)

            I = plt.imread(plotfn)
            I = I[:,:,0]
            sx,sy = (I*X).sum()/I.sum(), (I*Y).sum()/I.sum()
            #print sx,sy
            ok,ra,dec = anwcs_pixelxy2radec(plotwcs, sx, sy)

            #print era,edec
            #print ra,dec
            print('dRA,dDec', ra-era, dec-edec)
            self.assertTrue(abs(ra-era) < 1e-4)
            self.assertTrue(abs(dec-edec) < 1e-4)

if __name__ == '__main__':
    unittest.main()
